/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    ConditionalMutualInformation.java
 *    Copyright (C) 2005 Sam Danziger (Sam_Danziger@ieee.org)
 *
 */

package  weka.attributeSelection;

import  java.util.*;
import  weka.core.*;
import  weka.filters.unsupervised.attribute.Discretize;
import weka.filters.unsupervised.attribute.Normalize;
import  weka.filters.unsupervised.attribute.NumericToBinary;
import  weka.filters.Filter;

/** 
 * Class for Evaluating attributes individually by measuring information gain 
 * with respect to the class.
 *
 * Valid options are:<p>
 *
 * -M <br>
 * Treat missing values as a seperate value. <br>
 *
 * -B <br>
 * Just binarize numeric attributes instead of properly discretizing them. <br>
 *
 * @author Sam Danziger (Sam_Danziger@ieee.org)
 * @version $Revision: 1.0 $
 * @version $Revision: 1.1 $
 */
public class ConditionalMutualInformationAttributeEval
  extends AttributeEvaluator
  implements OptionHandler {

  /** Treat missing  */
  private int m_numFeatures;

  /** Just binarize numeric attributes */
  private boolean m_Binarize;

  /** The conditional mutual information gain for each attribute */
  private double[] m_InfoGains;

  /**
   * Returns a string describing this attribute evaluator
   * @return a description of the evaluator suitable for
   * displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return "ConditionalMutualInformation :\n\nEvaluates the worth of an attribute "
      +"by measuring the information gain with respect to the class and "
	  +"previously selected features.\n\n"
      +"ConditionalMutualInformation(Class,Attribute1,Attribute2) = H(Class|Attribute2) - H(Class | Attribute1,Attribute2).\n";
  }

  /**
   * Constructor
   */
  public ConditionalMutualInformationAttributeEval () {
    resetOptions();
  }

  /**
   * Returns an enumeration describing the available options.
   * @return an enumeration of all the available options.
   **/
  public Enumeration listOptions () {
    Vector newVector = new Vector(2);
    newVector.addElement(new Option("\tselect top A attributes rather than ranking every one\n" 
				    + "value.", "A", 0, "-A"));
    newVector.addElement(new Option("\tjust binarize numeric attributes instead\n " 
				    +"\tof properly discretizing them.", "B", 0, 
				    "-B"));
    return  newVector.elements();
  }


  /**
   * Parses a given list of options. <p>
   *
   * Valid options are:<p>
   *
   * -A <numFeatures> <br>
   * Set the number of attributes to select. <br>
   *
   * -B <br>
   * Just binarize numeric attributes instead of properly discretizing them. <br>
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   *
   **/
  public void setOptions (String[] options)
    throws Exception {

    resetOptions();
    
    String numFeatures = Utils.getOption('A', options);
    if (numFeatures.length() != 0) {
      setNumFeatures(Integer.parseInt(numFeatures));
    } else {
      setNumFeatures(Integer.MAX_VALUE);
    }

    setBinarizeNumericAttributes(Utils.getFlag('B', options));
  }


  /**
   * Gets the current settings of WrapperSubsetEval.
   *
   * @return an array of strings suitable for passing to setOptions()
   */
  public String[] getOptions () {
    String[] options = new String[2];
    int current = 0;

    options[current++] = "-A"; options[current++] = "" + getNumFeatures();
    if (getBinarizeNumericAttributes()) {
      options[current++] = "-B";
    }

    while (current < options.length) {
      options[current++] = "";
    }

    return  options;
  }

  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String binarizeNumericAttributesTipText() {
    return "Just binarize numeric attributes instead of properly discretizing them.";
  }

  /**
   * Binarize numeric attributes.
   *
   * @param b true=binarize numeric attributes
   */
  public void setBinarizeNumericAttributes (boolean b) {
    m_Binarize = b;
  }


  /**
   * get whether numeric attributes are just being binarized.
   *
   * @return true if missing values are being distributed.
   */
  public boolean getBinarizeNumericAttributes () {
    return  m_Binarize;
  }

  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String missingMergeTipText() {
    return "Distribute counts for missing values. Counts are distributed "
      +"across other values in proportion to their frequency. Otherwise, "
      +"missing is treated as a separate value.";
  }

  /**
   * distribute the counts for missing values across observed values
   *
   * @param b true=distribute missing values.
   */
  public void setNumFeatures (int n) {
    m_numFeatures = n;
  }


  /**
   * get whether missing values are being distributed or not
   *
   * @return true if missing values are being distributed.
   */
  public int getNumFeatures () {
    return  m_numFeatures;
  }


  /**
   * Initializes an information gain attribute evaluator.
   * Discretizes all attributes that are numeric.
   *
   * @param data set of instances serving as training data 
   * @exception Exception if the evaluator has not been 
   * generated successfully
   */
  public void buildEvaluator (Instances data)
    throws Exception {
/***/
  	if (false) {
  		throw  new Exception("Made it to the begining!");
  	}
/***/  	
    if (data.checkForStringAttributes()) {
      throw  new UnsupportedAttributeTypeException("Can't handle string attributes!");
    }
    
    int classIndex = data.classIndex();
    if (data.attribute(classIndex).isNumeric()) {
      throw  new Exception("Class must be nominal!");
    }
    int numInstances = data.numInstances();
    int numAttributes = data.numAttributes();
	
    int numBins=0;
    if (!m_Binarize) {
      numBins=10;
	  /*Unsupervised discritization has problems if 
	   * values are too close.  Fix this by normalizing
	   * each attribute first. 
	   *  Sam Danziger 1/07/06
	   */
      Normalize normTransform = new Normalize();
      normTransform.setInputFormat(data);
      data = Filter.useFilter(data, normTransform);

      Discretize disTransform = new Discretize();
      disTransform.setBins(numBins);
      disTransform.setInputFormat(data);
      data = Filter.useFilter(data, disTransform);
    } else {
      numBins=2;
      NumericToBinary binTransform = new NumericToBinary();
      binTransform.setInputFormat(data);
      data = Filter.useFilter(data, binTransform);
    }      
    
    // Convert the Instances into a data table
    int[][] dataTable = new int[numAttributes][numInstances];
    
    for (int i=0; i < numInstances; i++) {
	    Instance curInstance=data.instance(i);
    	for (int j=0; j < numAttributes; j++) {
    		dataTable[j][i]=(int)(curInstance.value(j));
    	}
    }
    
    // Calculate the unique values in each feature
    Object [] bins = new Object[numAttributes];
    for (int i=0; i< numAttributes; i++){
    	bins[i]=unique(dataTable[i],numBins);
    }
    
    // Step 1: Mutual Information
    //System.err.println("Step 1: Mutual Information");
    double [] mi = new double[numAttributes];

    double classEntropy =  calculateEntropy(dataTable[classIndex],numBins);
    for (int i = 0; i < data.numAttributes(); i++) {
      if (i != classIndex) {
      	double h = calculateConditionalEntropy(dataTable[classIndex],dataTable[i],numBins);
      	mi[i] = classEntropy - h;
      }
    }

    // Step 2: Maximum Conditional Mutual Information  
    if (m_numFeatures > numAttributes) {m_numFeatures=numAttributes;}
    Object [] theChosen=fastCMIM(mi, dataTable, numBins, bins, classIndex);
    int [] chosen=(int [])theChosen[0];
    double [] chosenMI=(double [])theChosen[1];
       
    
    //Step 3: Store the new values
    //System.err.println("Step 3: Store the new values");
    m_InfoGains = new double[data.numAttributes()];
	for (int i = 0; i < data.numAttributes(); i++) {
        if (i != classIndex) {
        	m_InfoGains[chosen[i]]=chosenMI[i];
        }
	}
   }
  
/***/
  
  /**
   * Return the index of the maximum value
   */
  private int findMaxIndex(double [] vector){
  	double curMax=0;
  	int maxIdx=-1;
  	
  	for (int i=0; i< vector.length; i++) {
  		if (vector[i]>curMax){
  			curMax=vector[i];
  			maxIdx=i;
  		}
  	}
  	
  	return maxIdx;
  }
  
  /**
   * Calculate the Conditional Mutual Information (CMIM)as follows
   *   I(U,Vn|Vm) = H(U,Vm)-H(Vm)-H(U,Vn,Vm)+H(Vn,Vm)
   *   Uses the most naive (slowest) method possible 
   * @param U the class vector
   * @param Vn the new attribute vector
   * @param Vm the attribute vector already chosen
   */
  protected double calculateCMIM_1(int[] U, int[] Vn, int[] Vm, int numBins) {
  	double cmim=0;
  	
  	double h1=calculateEntropy(U,Vm,numBins);
  	double h2=calculateEntropy(Vm,numBins);
  	double h3=calculateEntropy(U,Vn,Vm,numBins);
  	double h4=calculateEntropy(Vn,Vm,numBins);
  	
  	cmim=h1-h2-h3+h4;
  	
  	return cmim;
  }
  
  protected double calculateCMIM_1(int[] U, int[] Vn, int[] Vm, int[] binsU, int[] binsVn, int[] binsVm) {
  	double cmim=0;
  	
  	double h1=calculateEntropy(U,Vm,binsU,binsVm);
  	double h2=calculateEntropy(Vm,binsVm);
  	double h3=calculateEntropy(U,Vn,Vm,binsU,binsVn,binsVm);
  	double h4=calculateEntropy(Vn,Vm,binsVn,binsVm);
  	
  	cmim=h1-h2-h3+h4;
  	
  	return cmim;
  }
  
  /**
   * Calculate the Conditional Mutual Information (CMIM)as follows
   *   I(U,Vn|Vm) = H(U,Vm)-H(Vm)-H(U,Vn,Vm)+H(Vn,Vm)
   *   Optimized to eliminate unecessary calculation. 
   *   BROKEN:  Needs to be redirived
   * @param U the class vector
   * @param Vn the new attribute vector
   * @param Vm the attribute vector already chosen
   */
  protected double calculateCMIM_2(int[] U, int[] Vn, int[] Vm, int numBins) {
  	double cmim=0;
  	
  	int numAttributes=U.length;
  	
  	/*Temporary: calculate the exact size of bins*/
  	int [] binsU=unique(U,numBins);
  	int [] binsVn=unique(Vn,numBins);
  	int [] binsVm=unique(Vm,numBins);
  	
  	
  	for (int kIdx=0; kIdx<binsU.length; kIdx++){  //The number of classes in U
  		int k = binsU[kIdx];
  		for (int jIdx=0; jIdx<binsVn.length; jIdx++){  //The number of bins in Vn
  			int j = binsVn[jIdx];
  			for (int iIdx=0; iIdx<binsVm.length; iIdx++){ //The number of bins in Vm
  				int i = binsVm[iIdx];
  				
  	  			int countUVnVm=0;
  	  			int countUVn=0;
  	  			
  	  			int countU=0;
  	  			int countUVm=0;
  	    	  			
  				for (int n=0; n<numAttributes; n++){

  					//Calculate p(k,j,i) / P(U,Vn,Vm)
  					if(Vm[n]==i & Vn[n]==j & U[n]==k ){
  						countUVnVm++;
  					}

  					//Calculate p(k,i) / P(U,Vm)
  					/*if(j==0 & Vm[n]==i & U[n]==k ){
  						countUVm++;
  					}*/
  					if(Vm[n]==i & U[n]==k ){
  						countUVn++;
  					}
  				}//n
  				double pUVm=(double)countUVm/(double)numAttributes;
  				//double pVm=(double)countVm/(double)numAttributes;
  				double pUVnVm=(double)countUVnVm/(double)numAttributes;
  				//double pVnVm=(double)countVnVm/(double)numAttributes;
  				
  				cmim=cmim-lnFunc2(pUVnVm,pUVm);
  			}//i
  		}//j
  	}//k
  	
  	for (int kIdx=0; kIdx<binsU.length; kIdx++){  //The number of classes in U
  		int k = binsU[kIdx];
		for (int iIdx=0; iIdx<binsVm.length; iIdx++){ //The number of bins in Vm
			int i = binsVm[iIdx];
  				
  			int countU=0;
  			int countUVm=0;
  	    	  			
			for (int n=0; n<numAttributes; n++){

				//Calculate p(k,j,i) / P(U,Vn,Vm)
				if(Vm[n]==i & U[n]==k ){
					countUVm++;
				}

				//Calculate p(k,i) / P(U,Vm)
				if(U[n]==k){
					countU++;
				}
			}//n
			double pU=(double)countU/(double)numAttributes;
			//double pVm=(double)countVm/(double)numAttributes;
			double pUVm=(double)countUVm/(double)numAttributes;
			//double pVnVm=(double)countVnVm/(double)numAttributes;
 				
			cmim=cmim+lnFunc2(pUVm,pU);
  		}//i
  	}//k
  	
  	return cmim;
  }
  
  /**
   * Create a list of unique values in array input 
   * @param input the list of values
   * @param numBins the list of possible values input may take
   * @return
   */
  private int[] unique(int []input, int numBins){
  	boolean [] exists=new boolean[numBins];
  	Arrays.fill(exists,false);
  	
  	int count=0;
  	
  	
  	for (int i=0; i<numBins; i++){
  		for (int j=0; j<input.length; j++){
  			if (input[j]==i){
  				count++;
  				exists[i]=true;
  				break;
  			}
  		}
  	}
  	
  	int [] values = new int[count];
  	int usedVals=0;
  	for (int i=0; i<numBins; i++){
  		if (exists[i]==true){
  			values[usedVals]=i;
  			usedVals++;
  		}
  	}
  	  	 	
  	return values;
  }
  
  /**
   * Help method for computing entropy.
   */
  private static double lnFunc(double num){
    
    // Constant hard coded for efficiency reasons
    if (num < 1e-6) {
      return 0;
    } else {
      return num * Math.log(num);
    }
  }
  
  /**
   * Help method for computing entropy.
   * Return num1*log(num1/num2);
   */
  private static double lnFunc2(double num1,double num2){
    
    // Constant hard coded for efficiency reasons
    if (num1 < 1e-6) {
      return 0;
    } else {
      return num1 * Math.log(num1/num2);
    }
  }
  
  /**
   * Implement the simple (slow) Conditional Mutual Information algorithm as
   * described in Fleuret, Francois. Journal of Machine Learning Research 5:1531-1555 (2004)
   * Returns an object array with [0] an int[] with the chosen values and
   * [1] a double[] with the chosen MI values;
   * @param mi
   * @param dataTable
   * @param numBins
   * @param classIndex
   * @return
   */
  private Object[] simpleCMIM(double [] mi, int [][] dataTable, int numBins, int classIndex) {
  	//System.err.println("Step 2: Maximum Conditional Mutual Information");
  	int numAttributes=dataTable.length;
  	
  	double [] curMI= new double[numAttributes]; //Called "s"
  	int [] chosen= new int[numAttributes]; //Called "nu"
  	double [] chosenMI = new double[numAttributes]; //Track the winning MI scores
  
  	for (int i=0; i<numAttributes; i++){
  		curMI[i]=mi[i];
  	}
  
  	for (int k=0; k<numAttributes; k++) {
  		if (k != classIndex) {
  			chosen[k]=findMaxIndex(curMI);
  			chosenMI[k]=curMI[chosen[k]];
  			for (int n=0; n<numAttributes; n++) {
  				if (n != classIndex) {
  					double newMI=calculateCMIM_1(dataTable[classIndex],dataTable[n],dataTable[chosen[k]],numBins);
  					if ( newMI < curMI[n]) {
  						curMI[n]=newMI;
  					}
  				}
  			}
  		}
  	}
  	
  	Object [] theChosen=new Object[2];
  	theChosen[0]=chosen;
  	theChosen[1]=chosenMI;
  	return theChosen;
  }
  
  /**
   * Implement the fast Conditional Mutual Information algorithm as
   * described in Fleuret, Francois. Journal of Machine Learning Research 5:1531-1555 (2004)
   * Returns an object array with [0] an int[] with the chosen values and
   * [1] a double[] with the chosen MI values;
   * @param mi
   * @param dataTable
   * @param numBins
   * @param classIndex
   * @return
   */
  private Object[] fastCMIM(double [] mi, int [][] dataTable, int numBins, Object [] bins,int classIndex) {
  	//System.err.println("Step 2: Maximum Conditional Mutual Information");
  	int numAttributes=dataTable.length;
  	
  	double [] curMI= new double[numAttributes]; //Called "ps"
  	int [] count= new int[numAttributes]; //Called "m"
  	int [] chosen= new int[numAttributes]; //Called "nu"
  	double [] chosenMI = new double[numAttributes]; //Track the winning MI scores
  
  	Arrays.fill(chosenMI,0);
  	Arrays.fill(count,0);
  	
  	for (int i=0; i<numAttributes; i++){
  		curMI[i]=mi[i];
  	}
  
  	//for (int k=0; k<numAttributes; k++) {
  	for (int k=0; k<m_numFeatures; k++) {
  	  	double curBest= 0; //Called "s*, the best up do date score this iteration"

  		if (k != classIndex) {
  			for (int n=0; n<numAttributes; n++) {
  				if (n != classIndex) {
  					while (curMI[n] > curBest & count[n] < k) {
  						int newIdx=chosen[count[n]];
  						//double newMI=calculateCMIM_1(dataTable[classIndex],dataTable[n],dataTable[newIdx],numBins);
  						double newMI=calculateCMIM_1(dataTable[classIndex],dataTable[n],dataTable[newIdx],(int [])(bins[classIndex]),(int [])(bins[n]),(int [])(bins[newIdx]));
  						//double newMI2=calculateCMIM_2(dataTable[classIndex],dataTable[n],dataTable[newIdx],numBins);
  						if ( newMI < curMI[n]) {
  	  						curMI[n]=newMI;
  	  					}
  						count[n]++;
  					}
  					if (curMI[n] > curBest){
  						curBest=curMI[n];
  						chosen[k]=n;
  						chosenMI[k]=curMI[chosen[k]];
  					}
  				}
  			}
  		}
  	}
  	
  	Object [] theChosen=new Object[2];
  	theChosen[0]=chosen;
  	theChosen[1]=chosenMI;
  	return theChosen;
  }  
  
  /**
   * Calculate the conditional entropy for two attributes 
   * H(attrVect1|attrVect2) = H(attrVect1,attrVect2)-H(attrVect2)
   * 
   * @param attrVect1 The first discritized attribute vector
   * @param attrVect2 The second discritized attribute vector
   * @param numBins   The number of nominalized valuesan attribute can take
   */
  protected double calculateConditionalEntropy( int[] attrVect1, int [] attrVect2, int numBins ) {
  	double jointEntropy=0; //H(attrVect1,attrVect2)
  	double entropy2=0;     //H(attrVect2)
  	int numInstances1=attrVect1.length;
  	//  int numInstances2=attrVect2.length;
  	//  Assert attrVect1.length == attrVect2.length
  	
    //  Count the instance values
	for (int i=0; i<numBins; i++){  //Track the bins for attrVect1
		for (int j=0; j<numBins; j++){ //Track the bins for attrVect2
  			int jointCount=0; //joint entropy counter
  			int count2=0;  //attrVect2 entropy counter
  	  	  	for (int k=0; k<numInstances1; k++){
  	  			int curVal1=attrVect1[k];
  	  			int curVal2=attrVect2[k];
  	  			if (curVal1==i & curVal2==j) {
  	  				jointCount++;
  	  			}
  	  			if (i==0 & curVal2==j){
  	  				count2++;
  	  			}
  	  	  	}//k
  	    	
  	  	  	//  Calculate the entropy
  	  	  	double jointPercent=(double)(jointCount)/(double)(numInstances1);
  	  	  	jointEntropy=jointEntropy-lnFunc(jointPercent);
  	  	  	
  	  	  	if (i==0) {
  	  	  		double percent2=(double)(count2)/(double)(numInstances1);
	  	  		entropy2=entropy2-lnFunc(percent2);
  	  	  	}
  		}//j
	}//i
  	
	double entropy=jointEntropy-entropy2;
  	return entropy;
  }
  
  /**
   * Caculate the entropy for a single attribute
   * 
   * @param attrVect The discritized attribute vector
   * @param numBins  The number of nominalized valuesan attribute can take
   */
  protected double calculateEntropy( int[] attrVect, int numBins ) {
  	double entropy=0;
  	int numInstances=attrVect.length;
  	  	
    //  Count the instance values
  	for (int i=0; i<numBins; i++){
  	  int count=0;
  	  for (int j=0; j<numInstances; j++){
  		int curVal=attrVect[j];
  		if (curVal==i) {
  		  count++;
  		}
  	  }
  	  
      //  Calculate the entropy
  	  double percent=(double)(count)/(double)(numInstances);
      entropy=entropy-lnFunc(percent);
  	}
  	
  	return entropy;
  }
  /**
   * Caculate the entropy for a single attribute
   * 
   * @param attrVect The discritized attribute vector
   * @param bins  An array with the unique values attrVect may take
   */
  protected double calculateEntropy( int[] attrVect, int[] bins ) {
  	double entropy=0;
  	int numInstances=attrVect.length;
  	  	
    //  Count the instance values
  	for (int iIdx=0; iIdx<bins.length; iIdx++){
  	  int i=bins[iIdx];
  	  int count=0;
  	  for (int j=0; j<numInstances; j++){
  		int curVal=attrVect[j];
  		if (curVal==i) {
  		  count++;
  		}
  	  }
  	  
      //  Calculate the entropy
  	  double percent=(double)(count)/(double)(numInstances);
      entropy=entropy-lnFunc(percent);
  	}
  	
  	return entropy;
  }
  
  /**
   * Caculate the joint entropy for two attributes
   * 
   * @param attrVect1 The first discritized attribute vector
   * @param attrVect2 The second discritized attribute vector
   * @param numBins   The number of nominalized valuesan attribute can take
   */
  protected double calculateEntropy( int[] attrVect1, int [] attrVect2, int numBins ) {
  	double entropy=0;
  	int numInstances1=attrVect1.length;
  	//  int numInstances2=attrVect2.length;
  	//  Assert attrVect1.length == attrVect2.length
  	
    //  Count the instance values
	for (int i=0; i<numBins; i++){  //Track the bins for attrVect1
  		for (int j=0; j<numBins; j++){ //Track the bins for attrVect2
  			int count=0;
  	  	  	for (int k=0; k<numInstances1; k++){
  	  			int curVal1=attrVect1[k];
  	  			int curVal2=attrVect2[k];
  	  			if (curVal1==i & curVal2==j) {
  	  				count++;
  	  			}
  	  	  	}//k
  	    	
  	  	  	//  Calculate the entropy
  	  	  	double percent=(double)(count)/(double)(numInstances1);
  	  	  	entropy=entropy-lnFunc(percent);
  		}//j
	}//i
  	
  	return entropy;
  }
  
  /**
   * Caculate the joint entropy for two attributes
   * 
   * @param attrVect1 The first discritized attribute vector
   * @param attrVect2 The second discritized attribute vector
   * @param bins1 An array with the unique values attrVect1 may take
   * @param bins2 An array with the unique values attrVect2 may take
   */
  protected double calculateEntropy( int[] attrVect1, int [] attrVect2, int [] bins1, int [] bins2 ) {
  	double entropy=0;
  	int numInstances1=attrVect1.length;
  	//  int numInstances2=attrVect2.length;
  	//  Assert attrVect1.length == attrVect2.length
  	
    //  Count the instance values
	for (int iIdx=0; iIdx<bins1.length; iIdx++){  //Track the bins for attrVect1
		int i=bins1[iIdx];
  		for (int jIdx=0; jIdx<bins2.length; jIdx++){ //Track the bins for attrVect2
  			int j=bins2[jIdx];
  			int count=0;
  	  	  	for (int k=0; k<numInstances1; k++){
  	  			int curVal1=attrVect1[k];
  	  			int curVal2=attrVect2[k];
  	  			if (curVal1==i & curVal2==j) {
  	  				count++;
  	  			}
  	  	  	}//k
  	    	
  	  	  	//  Calculate the entropy
  	  	  	double percent=(double)(count)/(double)(numInstances1);
  	  	  	entropy=entropy-lnFunc(percent);
  		}//j
	}//i
  	
  	return entropy;
  }
  
  /**
   * Caculate the joint entropy for three attributes
   * 
   * @param attrVect1 The first discritized attribute vector
   * @param attrVect2 The second discritized attribute vector
   * @param attrVect3 The third discritized attribute vector
   * @param numBins   The number of nominalized valuesan attribute can take
   */
  protected double calculateEntropy( int[] attrVect1, int [] attrVect2, int [] attrVect3, int numBins ) {
  	double entropy=0;
  	int numInstances1=attrVect1.length;
  	//  Assert attrVect1.length == attrVect2.length == attrVect2.length
  	
    //  Count the instance values
	for (int i=0; i<numBins; i++){  //Track the bins for attrVect1
  		for (int j=0; j<numBins; j++){ //Track the bins for attrVect2
  	  		for (int k=0; k<numBins; k++){ //Track the bins for attrVect3
  	  			int count=0;
  	  	  	  	for (int l=0; l<numInstances1; l++){
  	  	  			int curVal1=attrVect1[l];
  	  	  			int curVal2=attrVect2[l];
  	  	  			int curVal3=attrVect3[l];
  	  	  			if (curVal1==i & curVal2==j & curVal3==k) {
  	  	  				count++;
  	  	  			}
  	  	  	  	}//l
  	  	    	
  	  	  	  	//  Calculate the entropy
  	  	  	  	double percent=(double)(count)/(double)(numInstances1);
  	  	  	  	entropy=entropy-lnFunc(percent);
  	  		}//k
  		}//j
	}//i
  	
  	return entropy;
  }
  
  /**
   * Caculate the joint entropy for three attributes
   * 
   * @param attrVect1 The first discritized attribute vector
   * @param attrVect2 The second discritized attribute vector
   * @param attrVect3 The third discritized attribute vector
   * @param bins1 An array with the unique values attrVect1 may take
   * @param bins2 An array with the unique values attrVect2 may take
   * @param bins3 An array with the unique values attrVect3 may take
   */
  protected double calculateEntropy( int[] attrVect1, int [] attrVect2, int [] attrVect3, int [] bins1, int [] bins2, int [] bins3 ) {
  	double entropy=0;
  	int numInstances1=attrVect1.length;
  	//  Assert attrVect1.length == attrVect2.length == attrVect2.length
  	
    //  Count the instance values
	for (int iIdx=0; iIdx<bins1.length; iIdx++){  //Track the bins for attrVect1
		int i=bins1[iIdx];
  		for (int jIdx=0; jIdx<bins2.length; jIdx++){ //Track the bins for attrVect2
  			int j=bins2[jIdx];
  	  		for (int kIdx=0; kIdx<bins3.length; kIdx++){ //Track the bins for attrVect3
  	  			int k=bins3[kIdx];
  	  			int count=0;
  	  	  	  	for (int l=0; l<numInstances1; l++){
  	  	  			int curVal1=attrVect1[l];
  	  	  			int curVal2=attrVect2[l];
  	  	  			int curVal3=attrVect3[l];
  	  	  			if (curVal1==i & curVal2==j & curVal3==k) {
  	  	  				count++;
  	  	  			}
  	  	  	  	}//l
  	  	    	
  	  	  	  	//  Calculate the entropy
  	  	  	  	double percent=(double)(count)/(double)(numInstances1);
  	  	  	  	entropy=entropy-lnFunc(percent);
  	  		}//k
  		}//j
	}//i
  	
  	return entropy;
  }
  
  /**
   * Reset options to their default values
   */
  protected void resetOptions () {
    m_InfoGains = null;
    m_numFeatures = Integer.MAX_VALUE;
    m_Binarize = false;
  }


  /**
   * evaluates an individual attribute by measuring the amount
   * of information gained about the class given the attribute.
   *
   * @param attribute the index of the attribute to be evaluated
   * @exception Exception if the attribute could not be evaluated
   */
  public double evaluateAttribute (int attribute)
    throws Exception {

    return m_InfoGains[attribute];
  }

  /**
   * Describe the attribute evaluator
   * @return a description of the attribute evaluator as a string
   */
  public String toString () {
    StringBuffer text = new StringBuffer();

    if (m_InfoGains == null) {
      text.append("Conditional Mutual Information attribute evaluator has not been built");
    }
    else {
      text.append("\tContional Mutual Information Ranking Filter");
      if (m_numFeatures < Integer.MAX_VALUE) {
	text.append("\n\tSelecting "+m_numFeatures+" best attributes");
      }
      if (m_Binarize) {
	text.append("\n\tNumeric attributes are just binarized");
      }
    }
    
    text.append("\n");
    return  text.toString();
  }

  
  // ============
  // Test method.
  // ============
  /**
   * Main method for testing this class.
   *
   * @param argv the options
   */
  public static void main (String[] args) {
    try {
      System.out.println(AttributeSelection.
			 SelectAttributes(new ConditionalMutualInformationAttributeEval(), args));
    }
    catch (Exception e) {
      e.printStackTrace();
      System.out.println(e.getMessage());
    }
  }
}


